﻿' SkeletalMesh extractor for Unreal 2 E3 Alpha

Imports UPLib

Public Class Form1

    Dim pkg As Package

    Private Sub btnOpen_Click(sender As System.Object, e As System.EventArgs) Handles btnOpen.Click
        If OpenDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            pkg = New Package(OpenDlg.FileName)
        End If
    End Sub

    Private Sub btnExportAll_Click(sender As System.Object, e As System.EventArgs) Handles btnExportAll.Click
        If FolderDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            Dim path = FolderDlg.SelectedPath

            ExportProgressBar.Value = 0
            ExportProgressBar.Maximum = pkg.ExportTable.FindAll(Function(x) x.ClassNameStr = "SkeletalMesh").Count
            For i = 0 To pkg.ExportTable.Count - 1
                Dim obj = pkg.ExportTable(i)
                Select Case obj.ClassNameStr
                    Case "SkeletalMesh"
                        Dim mesh As New SkeletalMesh(obj)
                        mesh.Save(path)
                        ExportProgressBar.Value += 1
                        ExportProgressBar.Refresh()
                End Select
            Next
            ExportProgressBar.Value = 0
            SetStatus("")
            MsgBox("Finished.")
        End If
    End Sub

    Sub SetStatus(text As String)
        labStatus.Text = text
        labStatus.Refresh()
    End Sub

End Class

Public Class LodMesh
    Inherits Mesh

    Public CollapsePointThus As New List(Of UShort)
    Public FaceLevel As New List(Of UShort)
    Public Faces As New List(Of MeshFace)
    Public CollapseWedgeThus As New List(Of UShort)
    Public Wedges As New List(Of MeshWedge)
    Public Materials As New List(Of MeshMaterial)
    Public SpecialFaces As New List(Of MeshFace)
    Public ModelVerts, SpecialVerts As Integer
    Public MeshScaleMax As Single
    Public LODHysteresis, LODStrength As Single
    Public LODMinVerts As Integer
    Public LODMorph, LODZDisplace As Single
    Public RemapAnimVerts As New List(Of UShort)
    Public OldFrameVerts As Integer
End Class

Public Structure MeshFace
    Dim iWedge0, iWedge1, iWedge2, MaterialIndex As UShort
End Structure

Public Structure MeshWedge
    Dim iVertex As UShort
    Dim TexUV As MeshUV
End Structure

Public Structure MeshUV
    Dim U, V As Byte
End Structure

Public Structure MeshMaterial
    Dim PolyFlags As UInteger
    Dim TextureIndex As Integer
End Structure

Public Class SkeletalMesh
    Inherits LodMesh

    Public Obj As ExportTableItem

    Public ExtWedges As New List(Of MeshExtWedge)
    Public Points As New List(Of Vector)

    Public RefSkeleton As New List(Of MeshBone)
    Public BoneWeightIdx As New List(Of VBoneInfIndex)
    Public BoneWeights As New List(Of VBoneInfluence)
    Public LocalPoints As New List(Of Integer)
    Public SkeletalDepth As Integer
    Public DefaultAnimation As Integer
    Public WeaponBoneIndex As Integer
    Public WeaponAdjust As Coords

    Sub New(obj As ExportTableItem)
        Form1.SetStatus("Parsing mesh...")
        Dim r = obj.Data.R
        Me.Obj = obj

        obj.Data.Position = 0
        obj.GetProperties()
        obj.Data.Position += 45

        Dim VertArraySize = r.ReadCompact()
        Assert(VertArraySize = 0)
        VertData = r.ReadInts(VertArraySize)

        Dim numTris = r.ReadCompact()
        Assert(numTris = 0)
        ReDim Tris(-1)

        Dim numTextures = r.ReadCompact()
        Assert(numTextures > 0 And numTextures < 256)
        For i = 0 To numTextures - 1
            Textures.Add(r.ReadCompact())
        Next

        Scale = r.ReadVector
        Origin = r.ReadVector
        RotOrigin = r.ReadRotator

        Dim NumCollapsePointThus = r.ReadCompact()
        Assert((VertArraySize Mod NumCollapsePointThus) = 0)
        For i = 0 To NumCollapsePointThus - 1
            CollapsePointThus.Add(r.ReadUInt16)
        Next

        Dim numFaceLevel = r.ReadCompact()
        Assert(numFaceLevel >= 0)
        For i = 0 To numFaceLevel - 1
            FaceLevel.Add(r.ReadUInt16)
        Next

        Dim numFaces = r.ReadCompact()
        Assert(numFaces >= 0)
        For i = 0 To numFaces - 1
            Dim face As New MeshFace
            face.iWedge0 = r.ReadUInt16
            face.iWedge1 = r.ReadUInt16
            face.iWedge2 = r.ReadUInt16
            face.MaterialIndex = r.ReadUInt16
            Faces.Add(face)

            Assert(face.iWedge0 < 32768)
            Assert(face.iWedge1 < 32768)
            Assert(face.iWedge2 < 32768)
            Assert(face.MaterialIndex < 32768)
        Next

        Dim numCollapseWedgeThus = r.ReadCompact()
        Assert(numCollapseWedgeThus >= 0)
        For i = 0 To numCollapseWedgeThus - 1
            CollapseWedgeThus.Add(r.ReadUInt16)
        Next

        Dim numWedges = r.ReadCompact()
        Assert(numWedges >= 0)
        Assert(numWedges = numCollapseWedgeThus)
        For i = 0 To numWedges - 1
            Dim w As MeshWedge
            w.iVertex = r.ReadUInt16
            w.TexUV.U = r.ReadSingle
            w.TexUV.V = r.ReadSingle
            Wedges.Add(w)
        Next

        Dim numMaterials = r.ReadCompact
        Assert(numMaterials >= 0 And numMaterials <= 200)
        For i = 0 To numMaterials - 1
            Dim m As MeshMaterial
            m.PolyFlags = r.ReadUInt32
            m.TextureIndex = r.ReadInt32
            Assert(m.TextureIndex >= 0 And m.TextureIndex <= 200) ' And m.TextureIndex < numTextures)
            Materials.Add(m)
        Next

        For i = 0 To Faces.Count - 1
            Assert(Faces(i).MaterialIndex < numMaterials)
        Next

        'The following is a guess...
        MeshScaleMax = r.ReadSingle
        LODHysteresis = r.ReadSingle
        LODStrength = r.ReadSingle
        LODMinVerts = r.ReadInt32
        LODMorph = r.ReadSingle
        LODZDisplace = r.ReadSingle

        Assert(MeshScaleMax < 200)
        Assert(LODHysteresis < 200)
        Assert(LODStrength < 200)
        Assert(LODMinVerts < 200)
        Assert(LODMorph < 200)
        Assert(LODZDisplace < 200)

        Dim numPoints = r.ReadCompact
        Assert(numPoints = NumCollapsePointThus)
        Assert(numPoints > 0 And numPoints < 32768)
        For i = 0 To numPoints - 1
            Points.Add(r.ReadVector())
            Assert(Not Points(i).IsWeird)
        Next

        Dim numRefSkeleton = r.ReadCompact
        Assert(numRefSkeleton >= 0 And numRefSkeleton < 32768)
        For i = 0 To numRefSkeleton - 1
            Dim mb As MeshBone
            mb.Name = r.ReadCompact
            mb.Flags = r.ReadUInt32
            mb.BonePos.Orientation.X = r.ReadSingle
            mb.BonePos.Orientation.Y = r.ReadSingle
            mb.BonePos.Orientation.Z = r.ReadSingle
            mb.BonePos.Orientation.W = r.ReadSingle
            mb.BonePos.Position = r.ReadVector
            mb.BonePos.Length = r.ReadSingle
            mb.BonePos.XSize = r.ReadSingle
            mb.BonePos.YSize = r.ReadSingle
            mb.BonePos.ZSize = r.ReadSingle
            mb.NumChildern = r.ReadInt32
            mb.ParentIndex = r.ReadInt32
            RefSkeleton.Add(mb)

            Assert(mb.Name >= 0 And mb.Name < obj.Pkg.NameTable.Count)
            Assert(Not IsNumWeird(mb.BonePos.Orientation.X))
            Assert(Not IsNumWeird(mb.BonePos.Orientation.Y))
            Assert(Not IsNumWeird(mb.BonePos.Orientation.Z))
            Assert(Not IsNumWeird(mb.BonePos.Orientation.W))
            Assert(Not mb.BonePos.Position.IsWeird)
            Assert(Not IsNumWeird(mb.BonePos.Length))
            Assert(Not IsNumWeird(mb.BonePos.XSize))
            Assert(Not IsNumWeird(mb.BonePos.YSize))
            Assert(Not IsNumWeird(mb.BonePos.ZSize))
            Assert(mb.NumChildern >= 0 And mb.NumChildern < 32768)
            Assert(mb.ParentIndex >= 0 And mb.ParentIndex < 32768)
        Next

        Dim numBoneWeightIdx = r.ReadCompact
        Assert(numBoneWeightIdx = numRefSkeleton)
        For i = 0 To numBoneWeightIdx - 1
            Dim bwi As VBoneInfIndex
            bwi.WeightIndex = r.ReadUInt16
            bwi.Number = r.ReadUInt16
            bwi.DetailA = r.ReadUInt16
            bwi.DetailB = r.ReadUInt16
            BoneWeightIdx.Add(bwi)
        Next

        Dim numBoneWeights = r.ReadCompact
        For i = 0 To numBoneWeights - 1
            Dim bw As VBoneInfluence
            bw.PointIndex = r.ReadUInt16
            bw.BoneWeight = r.ReadUInt16
            BoneWeights.Add(bw)

            Assert(bw.PointIndex < numPoints)
            Assert(bw.BoneWeight > 0)
        Next

        For Each bwi In BoneWeightIdx
            Assert(bwi.WeightIndex + bwi.Number <= numBoneWeights)
        Next

        DefaultAnimation = r.ReadCompact
        If DefaultAnimation <> 0 Then
            Assert(obj.Pkg.ExportTable(DefaultAnimation - 1).ClassNameStr = "MeshAnimation")
        End If

        r.ReadInt32() 'SkeletalDepth? e.g. 14
        WeaponBoneIndex = r.ReadInt32()

        Assert(WeaponBoneIndex < RefSkeleton.Count)

        WeaponAdjust.Origin = r.ReadVector
        WeaponAdjust.XAxis = r.ReadVector
        WeaponAdjust.YAxis = r.ReadVector
        WeaponAdjust.ZAxis = r.ReadVector

        Assert(Not WeaponAdjust.Origin.IsWeird)
        Assert(Not WeaponAdjust.XAxis.IsWeird)
        Assert(Not WeaponAdjust.YAxis.IsWeird)
        Assert(Not WeaponAdjust.ZAxis.IsWeird)

        If obj.Data.Position <> obj.Data.Size Then
            Throw New Exception("Serial size mismatch")
        End If
    End Sub

    Overloads Sub Save(rootDir As String) 'as .psk + .psa
        Dim anim As MeshAnimation = Nothing
        If DefaultAnimation <> 0 Then
            anim = New MeshAnimation(Obj.Pkg.ExportTable(DefaultAnimation - 1))
        End If

        Form1.SetStatus("Saving mesh...")
        Using w As New UnBinaryWriter(New IO.FileStream(IO.Path.Combine(rootDir, Obj.ObjectNameStr & ".psk"), IO.FileMode.Create))
            w.WriteFixedString("ACTRHEAD", 20)
            w.WriteInt32(0) 'Type flag(???)
            w.WriteInt32(0) 'Data Size
            w.WriteInt32(0) 'Data Count

            w.WriteFixedString("PNTS0000", 20)
            w.WriteInt32(0)
            w.WriteInt32(12)
            w.WriteInt32(Points.Count)
            For Each pt In Points
                w.WriteVector(pt)
            Next

            w.WriteFixedString("VTXW0000", 20)
            w.WriteInt32(0)
            w.WriteInt32(16)
            w.WriteInt32(Wedges.Count)
            For Each we In Wedges
                w.WriteUInt32(we.iVertex) 'What's the difference between formats?
                w.WriteSingle(we.TexUV.U / 256) '/ 256?
                w.WriteSingle(1 - we.TexUV.V / 256)
                w.WriteUInt32(0) 'Material index???
            Next

            w.WriteFixedString("FACE0000", 20)
            w.WriteInt32(0)
            w.WriteInt32(12)
            w.WriteInt32(Faces.Count)
            For Each fc In Faces
                w.WriteUInt16(fc.iWedge0)
                w.WriteUInt16(fc.iWedge1)
                w.WriteUInt16(fc.iWedge2)
                w.WriteByte(fc.MaterialIndex)
                w.WriteByte(0) 'Auxiliary material index
                w.WriteInt32(0) 'Smoothing groups
            Next

            w.WriteFixedString("MATT0000", 20)
            w.WriteInt32(0)
            w.WriteInt32(88)
            w.WriteInt32(Materials.Count)
            For i = 0 To Materials.Count - 1
                w.WriteFixedString("Material" & i, 64) '???
                w.WriteInt32(Materials(i).TextureIndex)
                w.WriteUInt32(Materials(i).PolyFlags)
                w.WriteInt32(0) 'Auxiliary material (reserved)
                w.WriteUInt32(0) 'Auxiliary flags (reserved)
                w.WriteUInt32(0) 'LOD bias (unused)
                w.WriteUInt32(0) 'LOD style (unused)
            Next

            w.WriteFixedString("REFSKELT", 20)
            w.WriteInt32(0)
            w.WriteInt32(120)
            w.WriteInt32(RefSkeleton.Count)
            For Each mb In RefSkeleton
                w.WriteFixedString(Obj.Pkg.GetName(mb.Name), 64)
                w.WriteUInt32(mb.Flags)
                w.WriteInt32(mb.NumChildern)
                w.WriteInt32(mb.ParentIndex)
                w.WriteSingle(mb.BonePos.Orientation.X)
                w.WriteSingle(mb.BonePos.Orientation.Y)
                w.WriteSingle(mb.BonePos.Orientation.Z)
                w.WriteSingle(mb.BonePos.Orientation.W)
                w.WriteVector(mb.BonePos.Position)
                w.WriteSingle(mb.BonePos.Length)
                w.WriteSingle(mb.BonePos.XSize)
                w.WriteSingle(mb.BonePos.YSize)
                w.WriteSingle(mb.BonePos.ZSize)
            Next

            w.WriteFixedString("RAWWEIGHTS", 20)
            w.WriteInt32(0)
            w.WriteInt32(12)
            Dim numBoneWeights As Integer
            For Each bwi In BoneWeightIdx
                numBoneWeights += bwi.Number
            Next
            w.WriteInt32(numBoneWeights)
            'BoneWeightIdx.Count = bone count, so they must be per-bone
            For iBone = 0 To BoneWeightIdx.Count - 1
                Dim bwi = BoneWeightIdx(iBone)

                For iWeight = bwi.WeightIndex To bwi.WeightIndex + bwi.Number - 1
                    w.WriteSingle(BoneWeights(iWeight).BoneWeight / 65535) '/65535?
                    w.WriteInt32(BoneWeights(iWeight).PointIndex)
                    w.WriteInt32(iBone)
                Next
            Next
        End Using

        Form1.SetStatus("Saving animations...")
        If anim IsNot Nothing Then
            Dim psaPath = IO.Path.Combine(rootDir, Obj.Pkg.ExportTable(DefaultAnimation - 1).ObjectNameStr & ".psa")
            If Not IO.File.Exists(psaPath) Then
                Using w As New UnBinaryWriter(New IO.FileStream(psaPath, IO.FileMode.Create))
                    w.WriteFixedString("ANIMHEAD", 20)
                    w.WriteInt32(0) 'Type flag(???)
                    w.WriteInt32(0)
                    w.WriteInt32(0)

                    w.WriteFixedString("BONENAMES", 20)
                    w.WriteInt32(0)
                    w.WriteInt32(120)
                    w.WriteInt32(anim.RefBones.Count)
                    For Each rb In anim.RefBones
                        Dim rb2 = rb 'shut up
                        Dim pskBone = RefSkeleton.Find(Function(x) x.Name = rb2.Name)

                        w.WriteFixedString(Obj.Pkg.GetName(rb.Name), 64)
                        w.WriteUInt32(rb.Flags)
                        w.WriteInt32(pskBone.NumChildern)
                        w.WriteInt32(rb.ParentIndex)
                        w.WriteSingle(pskBone.BonePos.Orientation.X)
                        w.WriteSingle(pskBone.BonePos.Orientation.Y)
                        w.WriteSingle(pskBone.BonePos.Orientation.Z)
                        w.WriteSingle(pskBone.BonePos.Orientation.W)
                        w.WriteVector(pskBone.BonePos.Position)
                        w.WriteInt32(0) 'should be empty?
                        w.WriteInt32(0)
                        w.WriteInt32(0)
                        w.WriteInt32(0)
                    Next

                    ' Skip the 'All' anim?
                    Dim iFirstSeq = 0
                    If Form1.cbNoBaseAnim.Checked And StrEq(Obj.Pkg.GetName(anim.AnimSeqs(0).Name), "Base") Then
                        iFirstSeq = 1
                    End If

                    w.WriteFixedString("ANIMINFO", 20)
                    w.WriteInt32(0)
                    w.WriteInt32(168)
                    w.WriteInt32(anim.AnimSeqs.Count - iFirstSeq)
                    Dim curFrame As Integer
                    For i = iFirstSeq To anim.AnimSeqs.Count - 1
                        Dim seq = anim.AnimSeqs(i)
                        Dim move = anim.Moves(i)
                        w.WriteFixedString(Obj.Pkg.GetName(seq.Name), 64)
                        w.WriteFixedString(Obj.Pkg.GetName(seq.Group), 64)
                        w.WriteInt32(move.BoneIndices.Count)
                        Assert(move.BoneIndices.Count = 0 Or move.BoneIndices.Count = anim.RefBones.Count)
                        w.WriteInt32(0) 'Root include (unused)
                        w.WriteInt32(0) 'Compression style (reserved)
                        w.WriteInt32(1) '??? (Key quotum)
                        w.WriteSingle(0) '??? (Key reduction)
                        w.WriteSingle(move.TrackTime)
                        w.WriteSingle(seq.Rate)
                        w.WriteInt32(move.StartBone)
                        w.WriteInt32(curFrame) '??? (First raw frame)
                        w.WriteInt32(seq.NumFrames)
                        curFrame += seq.NumFrames
                    Next
                    Dim totalFrames = curFrame

                    w.WriteFixedString("ANIMKEYS", 20)
                    w.WriteInt32(0)
                    w.WriteInt32(32)
                    w.WriteInt32(totalFrames * anim.RefBones.Count)
                    For i = iFirstSeq To anim.AnimSeqs.Count - 1
                        Dim seq = anim.AnimSeqs(i)
                        Dim move = anim.Moves(i)

                        Dim decompressedTracks As New List(Of AnalogTrack)
                        For Each at In move.AnimTracks
                            decompressedTracks.Add(at.Interpolate(seq.NumFrames))
                        Next

                        For iFrame = 0 To seq.NumFrames - 1
                            For Each track In decompressedTracks 'this is per-bone
                                w.WriteVector(track.GetKeyPos(iFrame))
                                w.WriteSingle(track.KeyQuat(iFrame).X)
                                w.WriteSingle(track.KeyQuat(iFrame).Y)
                                w.WriteSingle(track.KeyQuat(iFrame).Z)
                                w.WriteSingle(track.KeyQuat(iFrame).W)
                                w.WriteSingle(track.KeyTime(iFrame))
                            Next
                        Next
                    Next
                End Using
            End If
        End If
    End Sub
End Class

Public Structure MeshExtWedge
    Dim iVertex, Flags As UShort
    Dim TexUV As MeshFloatUV
End Structure

Public Structure MeshFloatUV
    Dim U, V As Single
End Structure

Public Structure MeshBone
    Dim Name As Integer
    Dim Flags As UInteger
    Dim BonePos As VJointPos
    Dim NumChildern As Integer
    Dim ParentIndex As Integer
End Structure

Public Structure VJointPos
    Dim Orientation As Quat
    Dim Position As Vector
    Dim Length, XSize, YSize, ZSize As Single
End Structure

Public Structure VBoneInfIndex
    Dim WeightIndex, Number, DetailA, DetailB As UShort
End Structure

Public Structure VBoneInfluence
    Dim PointIndex, BoneWeight As UShort
End Structure

Public Structure Coords
    Dim Origin, XAxis, YAxis, ZAxis As Vector
End Structure

Public Class MeshAnimation
    Public RefBones As New List(Of NamedBone)
    Public Moves As New List(Of MotionChunk)
    Public AnimSeqs As New List(Of MeshAnimSeq)

    Sub New(obj As ExportTableItem)
        Form1.SetStatus("Parsing animations...")
        Dim r = obj.Data.R

        obj.Data.Position = 0
        r.ReadProperties()

        Dim numRefBones = r.ReadCompact
        Assert(numRefBones >= 0)
        For i = 0 To numRefBones - 1
            Dim rb As NamedBone
            rb.Name = r.ReadCompact
            rb.Flags = r.ReadUInt32
            rb.ParentIndex = r.ReadInt32
            RefBones.Add(rb)

            Assert(rb.Name >= 0 And rb.Name < obj.Pkg.NameTable.Count)
        Next

        Dim numMoves = r.ReadCompact
        Assert(numMoves >= 0)
        For i = 0 To numMoves - 1
            Dim mc As New MotionChunk
            mc.RootSpeed3D = r.ReadVector
            mc.TrackTime = r.ReadSingle
            Assert(Not mc.RootSpeed3D.IsWeird)
            Assert(Not IsNumWeird(mc.TrackTime))
            mc.StartBone = r.ReadInt32
            Assert(mc.StartBone = 0)
            mc.Flags = r.ReadUInt32
            Dim numBoneIndices = r.ReadCompact
            Assert(numBoneIndices >= 0)
            For j = 0 To numBoneIndices - 1
                mc.BoneIndices.Add(r.ReadInt32)
                Assert(mc.BoneIndices(j) = j)
            Next
            Dim numAnimTracks = r.ReadCompact
            Assert(numAnimTracks >= 0)
            For j = 0 To numAnimTracks - 1
                Dim curTrack = ReadAnalogTrack(r)
                mc.AnimTracks.Add(curTrack)
                Assert(curTrack.KeyTime(curTrack.KeyTime.Count - 1) < mc.TrackTime)
            Next
            mc.RootTrack = ReadAnalogTrack(r)
            Assert(mc.RootTrack.KeyTime.Count = 0)
            Moves.Add(mc)
        Next

        Dim numAnimSeqs = r.ReadCompact
        Assert(numAnimSeqs >= 0)
        Assert(numAnimSeqs = Moves.Count)
        For i = 0 To numAnimSeqs - 1
            Dim mas As New MeshAnimSeq
            mas.Name = r.ReadCompact 'e.g. AnimDn_Fr01_FT
            Assert(mas.Name >= 0 And mas.Name < obj.Pkg.NameTable.Count)
            mas.Group = r.ReadCompact 'e.g. aiming
            Assert(mas.Group >= 0 And mas.Group < obj.Pkg.NameTable.Count)
            mas.BaseName = r.ReadCompact 'e.g. AnimDn_Fr. Not sure if this is BaseName.
            Assert(mas.BaseName >= 0 And mas.BaseName < obj.Pkg.NameTable.Count)
            mas.BaseNamePlus = r.ReadCompact 'e.g. AnimDn_Fr01
            Assert(mas.BaseNamePlus >= 0 And mas.BaseNamePlus < obj.Pkg.NameTable.Count)
            mas.Number = r.ReadCompact() 'None or 01, 02... or 1, 2...
            Assert(mas.Number >= 0 And mas.Number < obj.Pkg.NameTable.Count)
            mas.Suffix = r.ReadCompact 'e.g. FT
            Assert(mas.Suffix >= 0 And mas.Suffix < obj.Pkg.NameTable.Count)
            mas.Misc = r.ReadCompact() '???
            Assert(r.ReadInt32 = 0) 'unused StartFrame?
            mas.NumFrames = r.ReadInt32
            Assert(mas.NumFrames >= 0 And mas.NumFrames < 999999)
            Dim numNotifys = r.ReadCompact
            Assert(numNotifys >= 0)
            For j = 0 To numNotifys - 1
                Dim man As MeshAnimNotify
                man.Time = r.ReadSingle
                Assert(Not IsNumWeird(man.Time))
                man.Func = r.ReadCompact
                Assert(man.Func >= 0 And man.Func < obj.Pkg.NameTable.Count)
                mas.Notifys.Add(man)
            Next
            mas.Rate = r.ReadSingle
            Assert(Not IsNumWeird(mas.Rate))

            AnimSeqs.Add(mas)
        Next

        If obj.Data.Position <> obj.Data.Size Then
            Throw New Exception("Serial size mismatch")
        End If
    End Sub

    Private Shared Function ReadAnalogTrack(r As UnBinaryReader) As AnalogTrack
        Dim at As New AnalogTrack
        at.Flags = r.ReadUInt32
        Dim numKeyQuat = r.ReadCompact
        Assert(numKeyQuat >= 0)
        For i = 0 To numKeyQuat - 1
            Dim q As Quat
            q.X = r.ReadSingle
            q.Y = r.ReadSingle
            q.Z = r.ReadSingle
            q.W = r.ReadSingle
            at.KeyQuat.Add(q)

            Assert(Not IsNumWeird(q.X))
            Assert(Not IsNumWeird(q.Y))
            Assert(Not IsNumWeird(q.Z))
            Assert(Not IsNumWeird(q.W))
        Next

        Dim numKeyPos = r.ReadCompact
        Assert(numKeyPos >= 0)
        For i = 0 To numKeyPos - 1
            at.KeyPos.Add(r.ReadVector)
            Assert(Not at.KeyPos(i).IsWeird)
        Next

        Dim numKeyTime = r.ReadCompact
        Assert(numKeyTime >= 0)
        For i = 0 To numKeyTime - 1
            at.KeyTime.Add(r.ReadSingle)
            Assert(Not IsNumWeird(at.KeyTime(i)))
        Next

        Return at
    End Function
End Class

Public Structure NamedBone
    Dim Name As Integer
    Dim Flags As UInteger
    Dim ParentIndex As Integer
End Structure

Public Class MotionChunk
    Public RootSpeed3D As Vector
    Public TrackTime As Single
    Public StartBone As Integer
    Public Flags As UInteger
    Public BoneIndices As New List(Of Integer)
    Public AnimTracks As New List(Of AnalogTrack)
    Public RootTrack As AnalogTrack 'Always empty
End Class

Public Class AnalogTrack
    Public Flags As UInteger
    Public KeyQuat As New List(Of Quat)
    Public KeyPos As New List(Of Vector)
    Public KeyTime As New List(Of Single)

    Function GetKeyPos(num As Integer) As Vector
        If KeyPos.Count = 1 Then Return KeyPos(0)

        Return KeyPos(num)
    End Function

    Private Function GetInterpKeyPos(time As Single) As Vector
        If KeyPos.Count = 1 Then Return KeyPos(0)

        If time >= KeyTime(KeyTime.Count - 1) Then
            Return KeyPos(KeyPos.Count - 1)
        End If

        'this is very unoptimized, but who cares
        Dim iFloor As Integer
        For i = KeyTime.Count - 2 To 0 Step -1
            If KeyTime(i) <= time Then
                iFloor = i
                Exit For
            End If
        Next

        Dim frac = (time - KeyTime(iFloor)) / (KeyTime(iFloor + 1) - KeyTime(iFloor))

        Return KeyPos(iFloor) + (KeyPos(iFloor + 1) - KeyPos(iFloor)) * frac
    End Function

    Private Function GetInterpKeyQuat(time As Single) As Quat
        If time >= KeyTime(KeyTime.Count - 1) Then
            Return KeyQuat(KeyPos.Count - 1)
        End If

        'this is very unoptimized, but who cares
        Dim iFloor As Integer
        For i = KeyTime.Count - 2 To 0 Step -1
            If KeyTime(i) <= time Then
                iFloor = i
                Exit For
            End If
        Next

        Dim frac = (time - KeyTime(iFloor)) / (KeyTime(iFloor + 1) - KeyTime(iFloor))

        Dim key1 = KeyQuat(iFloor)
        Dim key2 = KeyQuat(iFloor + 1)

        Dim result As Quat

        result.X = key1.X + (key2.X - key1.X) * frac
        result.Y = key1.Y + (key2.Y - key1.Y) * frac
        result.Z = key1.Z + (key2.Z - key1.Z) * frac
        result.W = key1.W + (key2.W - key1.W) * frac

        'Normalize it.
        Dim size = Math.Sqrt(result.X ^ 2 + result.Y ^ 2 + result.Z ^ 2 + result.W ^ 2)
        result.X /= size
        result.Y /= size
        result.Z /= size
        result.W /= size

        Return result
    End Function

    Function Interpolate(newFrameCount As Integer) As AnalogTrack
        Assert(newFrameCount >= KeyTime.Count)
        If KeyTime.Count = newFrameCount Then Return Me

        Dim result As New AnalogTrack

        result.Flags = Flags
        For i = 0 To newFrameCount - 1
            Dim time = i / newFrameCount
            result.KeyTime.Add(time)
            result.KeyPos.Add(GetInterpKeyPos(time))
            result.KeyQuat.Add(GetInterpKeyQuat(time))
        Next

        Return result
    End Function
End Class

Public Structure Quat
    Dim X, Y, Z, W As Single
End Structure

Public Class MeshAnimSeq
    Public Name, Group, BaseName, NumFrames As Integer
    Public BaseNamePlus, Suffix, Number, Misc As Integer 'unknown purpose of these
    Public Notifys As New List(Of MeshAnimNotify)
    Public Rate As Single
End Class

Public Structure MeshAnimNotify
    Dim Time As Single
    Dim Func As Integer
End Structure